/*
 * Copyright (c) 2006 Hypertriton, Inc. <http://www.hypertriton.com/>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <EXTERN.h>
#include <perl.h>

#include "mailprocd.h"

char saVersion[64];
SV *saInstance = NULL;

#undef SA_ASSERT
#define SA_ASSERT(expr) if (!(expr)) { return (-1); }
#define SA_ASSERT_SP(expr) if (!(expr)) { goto fail; }

/* Return the SpamAssassin version into a fixed-size buffer. */
int
SA_GetVersion(char *buf, size_t buf_len)
{
	dSP;
	int n;
	SV *sv;
	char *s;
	STRLEN slen;

	ENTER; SAVETMPS;
	PUSHMARK(SP);

	n = call_pv("Mail::SpamAssassin::Version", G_SCALAR);
	SPAGAIN;
	SA_ASSERT_SP(n == 1);
	sv = POPs;
	s = SvPV(sv, slen);
	strlcpy(buf, s, buf_len);
	buf[slen] = '\0';

	PUTBACK;
	FREETMPS; LEAVE;
	return (0);
fail:
	PUTBACK;
	FREETMPS; LEAVE;
	return (-1);
}

/* Return the SpamAssassin context object. */
static __inline__ SV *
SA_Instance(void)
{
	if (saInstance != NULL) {
		return (saInstance);
	}
	if ((saInstance = get_sv("Assassin", 0)) == NULL ||
	    SvTYPE(saInstance) != SVt_RV) {
		Fatal("Missing SA instance");
	}
	return (saInstance);
}

/* Parse a message and return a new Mail::SpamAssassin::Message object. */
int
SA_ParseMessage(SF_Message *msg)
{
	dSP;
	I32 n;
	SV *svText;

	SA_ASSERT(msg->text != NULL);
	SA_ASSERT(svText = newSVpvn(msg->text, msg->text_len+1));

	ENTER; SAVETMPS;
	PUSHMARK(SP);
	XPUSHs(SA_Instance());
	XPUSHs(sv_2mortal(svText));
	PUTBACK;
	n = call_method("parse", G_SCALAR);
	SPAGAIN;
	SA_ASSERT_SP(n == 1);
	SA_ASSERT_SP(msg->svParsed = newSVsv(POPs));
	PUTBACK; FREETMPS;
	LEAVE;
	return (0);
fail:
	PUTBACK; FREETMPS;
	LEAVE;
	return (-1);
}

/* Check a message and return its spam status. */
int
SA_CheckMessage(SF_Message *msg)
{
	dSP;
	I32 n;

	SA_ASSERT(msg->svParsed != NULL);

	/* Run check. */
	ENTER; SAVETMPS;
	PUSHMARK(SP);
	XPUSHs(SA_Instance());
	XPUSHs(msg->svParsed);
	PUTBACK;
	n = call_method("check", G_SCALAR);
	SPAGAIN;
	SA_ASSERT_SP(n == 1);
	if (msg->status == NULL) {
		if ((msg->status = malloc(sizeof(SF_MsgStatus))) == NULL) {
			goto fail;
		}
		SF_MessageInitStatus(msg->status);
	} else {
		if (msg->status->sv != NULL)
			sv_free(msg->status->sv);
	}
	if ((msg->status->sv = newSVsv(POPs)) == NULL) {
		goto fail_free;
	}
	PUTBACK;
	FREETMPS; LEAVE;

	/* Obtain score. */
	ENTER; SAVETMPS;
	PUSHMARK(SP);
	XPUSHs(msg->status->sv);
	PUTBACK;
	n = call_method("get_score", G_SCALAR);
	SPAGAIN;
	if (n != 1) { goto fail_free; }
	msg->status->score = (float)POPn;
	PUTBACK;
	FREETMPS; LEAVE;
	
	/* Obtain user-configured spam threshold. */
	ENTER; SAVETMPS;
	PUSHMARK(SP);
	XPUSHs(msg->status->sv);
	PUTBACK;
	n = call_method("get_required_score", G_SCALAR);
	SPAGAIN;
	if (n != 1) { goto fail_free; }
	msg->status->req_score = (float)POPn;
	PUTBACK;
	FREETMPS; LEAVE;
	msg->status->spam_status = (msg->status->score >
	                            msg->status->req_score);
	return (0);
fail_free:
	if (msg->status != NULL) {
		if (msg->status->sv != NULL) { sv_free(msg->status->sv); }
		free(msg->status);
		msg->status = NULL;
	}
fail:
	PUTBACK;
	FREETMPS; LEAVE;
	return (-1);
}

/* Rewrite a message, replacing the original text. */
int
SA_RewriteMessage(SF_Message *msg)
{
	dSP;
	I32 n;
	STRLEN nlen;
	char *msgNew, *s;

	ENTER; SAVETMPS;
	PUSHMARK(SP);
	XPUSHs(msg->status->sv);
	PUTBACK;
	n = call_method("rewrite_mail", G_SCALAR);
	SPAGAIN;
	SA_ASSERT_SP(n == 1);
	SA_ASSERT_SP(msg->svRewrite = newSVsv(POPs));
	s = SvPV(msg->svRewrite, nlen);

	if ((msgNew = malloc(nlen+1)) == NULL) {
		goto fail;
	}
	free(msg->text);
	msg->text = msgNew;
	memcpy(msg->text, s, nlen);
	msg->text[nlen] = '\0';
	msg->text_len = (size_t)nlen-1;

	PUTBACK; FREETMPS;
	LEAVE;
	return (0);
fail:
	PUTBACK; FREETMPS;
	LEAVE;
	return (-1);
}

