/*
 * Copyright (c) 2007 Hypertriton, Inc. <http://www.hypertriton.com/>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <EXTERN.h>
#include <perl.h>

#include <sys/mman.h>

#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <syslog.h>
#include <stdarg.h>

#include "mailprocd.h"

/* Initialize a table. */
void
table_init(struct table *ta)
{
	int i;

	ta->data = NULL;
	ta->data_len = 0;
	for (i = 0; i < TABLE_NBUCKETS; i++) {
		SLIST_INIT(&ta->buckets[i].ents);
	}
	SLIST_INIT(&ta->ents);
}

/* Release the resources allocated by a table. */
void
table_destroy(struct table *ta)
{
	struct table_ent *ent, *nent;
	int i;

	for (i = 0; i < TABLE_NBUCKETS; i++) {
		for (ent = SLIST_FIRST(&ta->buckets[i].ents);
		     ent != NULL;
		     ent = nent) {
			nent = SLIST_NEXT(ent, bents);
			if (ent->key != NULL) { free(ent->key); }
			if (ent->data != NULL) { free(ent->data); }
			free(ent);
		}
	}
	if (munmap(ta->data, ta->data_len) == -1) {
		Debug("munmap() failed on table!");
	}
	if (close(ta->fd) == -1) {
		Debug("close() failed on table!");
	}
}

static __inline u_int
table_hash(const char *key)
{
	u_int h;
	u_char *p;

	for (h = 0, p = (u_char *)key; *p != '\0'; p++) {
		h = 31 * h + *p;
	}
	return (h % TABLE_NBUCKETS);
}

char *
table_lookup(struct table *ta, const char *key)
{
	struct table_ent *ent;

	if ((ent = table_lookup_hash(ta, key, table_hash(key))) == NULL) {
		return (NULL);
	}
	return (ent->data);
}

/* Look up a named table entry by hash. */
struct table_ent *
table_lookup_hash(struct table *ta, const char *key, u_int h)
{
	struct table_ent *ent;

	SLIST_FOREACH(ent, &ta->buckets[h].ents, bents) {
		if (strcmp(key, ent->key) == 0)
			break;
	}
	if (ent == NULL) {
		SF_SetError("Element `%s' was not found in the table.", key);
	}
	return (ent);
}

/* Copy a named table entry field to a fixed-size buffer. */
int
table_lookup_copy(struct table *ta, const char *key, char *buf, size_t buf_len)
{
	struct table_ent *ent;

	if ((ent = table_lookup_hash(ta, key, table_hash(key))) != NULL) {
		if (strlcpy(buf, ent->data, buf_len) >= buf_len) {
			SF_SetError("String too large for buffer.");
			return (-1);
		}
		return (0);
	}
	return (-1);
}

/* Insert a new table entry. */
int
table_insert(struct table *ta, struct table_ent *ent)
{
	u_int h = table_hash(ent->key);
	struct table_ent *tent;

	if ((tent = table_lookup_hash(ta, ent->key, h)) != NULL) {
		SF_SetError("Element `%s' exists in table (ent=[%s:%s])", ent->key,
		    tent->key, tent->data);
		return (-1);
	}
	SLIST_INSERT_HEAD(&ta->buckets[h].ents, ent, bents);
	SLIST_INSERT_HEAD(&ta->ents, ent, ents);
	ta->buckets[h].nents++;
	return (0);
}

/* Remove a named table entry. */
int
table_remove(struct table *ta, const char *key)
{
	u_int h = table_hash(key);
	struct table_ent *ent;

	if ((ent = table_lookup_hash(ta, key, h)) == NULL) {
		SF_SetError("Element `%s' not found in table.", key);
		return (-1);
	}
	SLIST_REMOVE(&ta->buckets[h].ents, ent, table_ent, bents);
	SLIST_REMOVE(&ta->ents, ent, table_ent, ents);

	if (ent->data != NULL) { free(ent->data); }
	if (ent->key != NULL) { free(ent->key); }
	free(ent);
	return (0);
}

/*
 * Separate the key and data values of a line.
 *
 * XXX does not allow spaces in keys; data is bound.
 */
static int
table_parse_line(struct table *ta, char *s)
{
	char data_buf[4096];
	struct table_ent *nent;
	char *key, *p;
	int i;

	if ((key = strsep(&s, " \t")) == NULL || key[0] == '\0') {
		SF_SetError("Bad key");
		return (-1);
	}
	if ((nent = malloc(sizeof(struct table_ent))) == NULL) {
		SF_SetError("Out of memory");
		return (-1);
	}
	if ((nent->key = strdup(key)) == NULL) {
		free(nent);
		SF_SetError("Out of memory");
		return (-1);
	}
	data_buf[0] = '\0';
	i = 0;
	while ((p = strsep(&s, " ")) != NULL) {
		if (i > 0) {
			if (strlcat(data_buf, " ", sizeof(data_buf)) >=
			    sizeof(data_buf)) {
				SF_SetError("Key %s: Value too large", key);
				goto fail;
			}
		}
		if (strlcat(data_buf, p, sizeof(data_buf)) >=
		    sizeof(data_buf)) {
			SF_SetError("Key %s: Value too large", key);
			goto fail;
		}
		i++;
	}
	if (table_insert(ta, nent) == -1) {
		goto fail;
	}
	if ((nent->data = strdup(data_buf)) == NULL) {
		SF_SetError("Out of memory");
		goto fail;
	}
	return (0);
fail:
	free(nent->key);
	free(nent);
	return (-1);
}

/*
 * Load a table from the given data file. Leading spacing characters and
 * characters between '#' and end of line are ignored.
 */
struct table *
table_load(const char *path, int flags)
{
	struct table *ta;
	char *data, *p, *eol;
	size_t len;
	u_int nline = 0, nelements = 0, nerrors = 0;
	int fd;
	int open_fl, mmap_prot;

	if ((flags & TABLE_READ) && (flags & TABLE_WRITE)) {
		open_fl = O_RDWR;
		mmap_prot = PROT_READ|PROT_WRITE;
	} else if (flags & TABLE_WRITE) {
		open_fl = O_WRONLY;
		mmap_prot = PROT_WRITE;
	} else {
		open_fl = O_RDONLY;
		mmap_prot = PROT_READ;
	}
	if ((fd = open(path, open_fl)) == -1) {
		SF_SetError("%s: %s", path, strerror(errno));
		return (NULL);
	}
	if ((len = (size_t)lseek(fd, 0, SEEK_END)) == 0) {
		SF_SetError("%s: empty", path);
		return (NULL);
	}
	Debug("%s: Mapping database (%lu bytes)", path, (unsigned long)len);

	if ((data = mmap(NULL, len, PROT_READ|PROT_WRITE, MAP_FILE|MAP_PRIVATE,
	    fd, 0)) == MAP_FAILED) {
		SF_SetError("%s: mmap: %s", path, strerror(errno));
		return (NULL);
	}

	if ((ta = malloc(sizeof(struct table))) == NULL) {
		SF_SetError("Out of memory");
		return (NULL);
	}
	table_init(ta);
	ta->data = data;
	ta->data_len = len;
	ta->fd = fd;

	if (data != NULL) {
		for (p = &data[0]; *p != '\0'; p++) {
			while (isspace(*p)) {
				p++;
			}
			if (*p == '#') {
				while (*(++p) != '\n')
				    ;;
			}
			for (eol = p;
			     *eol != '\n' && *eol != '\0';
			     eol++) {
				if (*eol == '#') {
					for (eol--; isspace(*eol); eol--)
					    ;;
					eol++;
					break;
				}
			}
			*eol = '\0';
			if (p < eol) {
				char *sp;
				char *pdup;

				for (sp = &eol[-1]; isspace(*sp); sp--)
				    ;;
				sp[1] = '\0';
				if ((pdup = strdup(p)) == NULL) {
					SF_SetError("Out of memory");
					return (NULL);
				}
				if (table_parse_line(ta, pdup) == -1) {
					Debug("%s:%u: %s",
					    path, nline, SF_GetError());
					syslog(LOG_ERR, "%s:%u: %s",
					    path, nline, SF_GetError());
					nerrors++;
				} else {
					nelements++;
				}
				free(pdup);
			}
			p = eol;
			nline++;
		}
	}
	Debug("%s: %u elements, %u errors", path, nelements, nerrors);
	return (ta);
}

/* Save the contents of the given table to the named file. */
int
table_save(struct table *ta, const char *path, mode_t mode)
{
	struct table_ent *ent;
	FILE *f;

	if ((f = fopen(path, "w")) == NULL) {
		SF_SetError("%s: %s", path, strerror(errno));
		return (-1);
	}
	TABLE_FOREACH(ent, ta) {
		fputs(ent->key, f);
		fputc(' ', f);
		fputs(ent->data, f);
		fputc('\n', f);
	}
	if (fchmod(fileno(f), mode) == -1) {
		syslog(LOG_ERR, "chmod %s: %m", path);
	}
	fclose(f);
	return (0);
}

